#ifndef _EVENT_H
#define _EVENT_H

// Events are used to switch the controllers ON and OFF. The orders can be sent immediatly, using the 
// sendOrder function, or they can be scheduled to happen later, using the addEvent function
#include <WProgram.h>

#include "Clock.h"
#include "ChaconController.h"

ChaconController chacon = ChaconController(2, 17, 161, 52);

typedef struct
{
    const prog_uchar* name;    // Name of the device
    byte type;                 // Type of the device
    byte control;              // Type of control used on the device  
    byte image;                // First image associated with the device
    int pos_x;                 // X position of the device in the blueprint
    int pos_y;                 // Y position of the device in the blueprint
    byte data[4];              // Generic data, different for each device
} tDevice, *pDevice;


// This function always returns 1, which means the order of the devices will be the order
// they are created in the code. Changing this funcion we can order them differently
// For instance, we can order by device type, or alfabetically
int compareDevices(pDevice d1 , pDevice d2)
{
    return 1;
}

// Destroys a device, freeing it's memory
void destroyDevice(pDevice device)
{
    free(device);
}

// Ordered Linked List containing all the devices active in the system
OrderedList <tDevice> devices(compareDevices, destroyDevice); 


// Different types of control implemented
#define DEVICE_CONTROL_SHIFT_REGISTER      1
#define DEVICE_CONTROL_CHACON              2
#define DEVICE_CONTROL_IR                  3
#define DEVICE_CONTROL_DIRECT_PIN          4

typedef struct
{
    long time;
    byte control;
    byte controller_id;
    byte device_id;
    byte new_device_status;
    byte data;
} tEvent, *pEvent;

// Compares two events, orderine them by execution time
int compareEvents(pEvent e1 , pEvent e2)
{
    return (e1->time>e2->time ? 1 : (e1->time==e2->time ? 0 : -1) ); 
}

// Destroys an event, freeing it's memory
void destroyEvent(pEvent event)
{
    free(event);
}

// Ordered list containing all the events in the system
OrderedList <tEvent> events(compareEvents, destroyEvent);

// Creates a new event and adds it to the event list
void addEvent(long time, byte control, byte controller_id, byte data, byte device_id, byte new_device_status)
{
    pEvent new_event = (pEvent)malloc(sizeof(tEvent));
    new_event->time = time;
    new_event->control = control;
    new_event->controller_id = controller_id;
    new_event->data = data;
    new_event->device_id = device_id;
    new_event->new_device_status = new_device_status;
    events.addElement(new_event);
}

// Removes all the upcoming events of a specific device
void removeEvent(byte control, byte controller_id)
{
    byte dev_pos = 1;
    while (dev_pos <= events.count())
    {
        pEvent eve = events.getPosition(dev_pos);
        if(eve->control == control && eve->controller_id == controller_id)
            events.removeElement(dev_pos);
        else dev_pos++;    
    }    
}

// Returns the execution time for a specific event
long getEventTime(int i)
{
    return events.getPosition(i)->time;
}

// Returns the controller type for a specific event
byte getEventControl(int i)
{
    return events.getPosition(i)->control;
}

// Sends an order to a specific device
// The current orders only implement ON/OFF commands, but the data is already stored in a byte
// This means it is also possible in the future to implement orders passing numerical values,
// for instance to control a dimmer light, or the temperature of an AC
void sendOrder(byte dev_control, byte controller_id, byte dev_data, byte device_id, byte new_device_status)
{
    #ifdef DEBUG_EVENTS
    switch (dev_control)
    {
        case DEVICE_CONTROL_WIRED: Serial.print("WIRED:"); break;
        case DEVICE_CONTROL_CHACON: Serial.print("CHACON:"); break;
    }
    Serial.print(controller_id, DEC);
    Serial.print(":");
    Serial.println(dev_data, DEC);
    #endif    
    // To add new types of control, simply add a new #define in the top, and
    // add a new case statement here, calling the function that will activate
    // the new control control type to implement
    switch (dev_control)
    {
        case DEVICE_CONTROL_SHIFT_REGISTER: sr.setRegisterNow(controller_id, dev_data); break;
        case DEVICE_CONTROL_CHACON: chacon.sendCommand(controller_id, dev_data); break;
        case DEVICE_CONTROL_DIRECT_PIN: digitalWrite(controller_id, dev_data); break;
//      case YOUR_CONTROL_HERE: myControlSendCommand(byte controller_id, byte device_order); break;
    }
    // Sets the new status of the device
    pDevice dev = devices.getPosition(device_id);
    dev->data[0] = new_device_status;
}

// Checks if it's time to process any new events
void processEvents(long previous_time, long time)
{
    // If the previous clock measurement was after 11:59:50 and the current one is before
    // 0:00:10, it means the day has changed, so we need to adjust the time on the event
    // list. This happens when an event is queued from one day to the other. The value of 
    // the number of seconds will overflow, so when the day changes we need to adjust all 
    // of the remaining values
    if (previous_time > (CLOCK_MAX_DAY_SECONDS-10) && time < 10)
    {
        for (byte i=1; i<=events.count(); i++)
        {
            pEvent event = events.getPosition(i);
            #ifdef DEBUG_EVENTS
            Serial.print("Changing:");
            Serial.print(event->time);
            Serial.print("->");
            #endif
            event->time -= CLOCK_MAX_DAY_SECONDS;
            Serial.print(event->time);
        }        
    }    

    while (events.count()>0 && time>=getEventTime(1))
    {
        pEvent event = events.getPosition(1);
        #ifdef DEBUG_EVENTS
        Serial.print("Event: dev:"); Serial.print(event->controller_id, DEC); Serial.print(":");
        Serial.print('  data:'); Serial.println(events.getPosition(1)->data, DEC);
        #endif
        sendOrder(event->control, event->controller_id, event->data, event->device_id, event->new_device_status);
        events.removeElement(1);
    }
}
#endif



