#ifndef _SCHEDULE_H
#define _SCHEDULE_H

/* Schedules are used to trigger device actions in the future. These actions are what appears
   to the user on the scheduler page. These schedules may be saved in the Arduino EEprom, so that 
   they are loaded whenever Ardomotic starts. Each schedule will trigger one or more events,
   depending on the device type it refers to.
*/

#include <WProgram.h>
#include "Device.h"

#define ACTION_CHANGE_STATE 0
#define ACTION_MOVE_MOTOR 1
#define ACTION_MOVE_STEPPER 2
#define MAX_ACTION_TYPES 3

#define EEPROM_ACTION_ADDRESS 100

typedef struct
{
    long time;
    byte device;
    byte data[4];
} tAction, *pAction;

// Compares two actions, orderine them by execution time
int compareSchedules(pAction a1 , pAction a2)
{
    return (a1->time>a2->time ? 1 : (a1->time==a2->time ? 0 : -1) );
}

// Destroys an action, freeing it's memory
void destroySchedule(pAction action)
{
    free(action);
}

// Ordered Linked List containing all the actions active in the system
OrderedList <tAction> schedules(compareSchedules, destroySchedule);

// Next action to be executed on the current day 
// This number resets to 1 every day at midnight
byte nextAction = 1;    // Last action executed on the current day

// Returns the execution time for a specific action
long getActionTime(int i)
{
    return schedules.getPosition(i)->time;
} 

// Returns the device of a specific action
byte getScheduleDevice(int i)
{
    return schedules.getPosition(i)->device;
}

// Returns the execution command for a specific action
byte getScheduleCommand(int i)
{
    return schedules.getPosition(i)->data[0];
}

// Adds a new action to the actions list
// It does NOT save it to flash memory
void addSchedule(long time, byte device, byte command)
{
    #ifdef DEBUG_ACTIONS
    Serial.print("Adding action: time:");
    Serial.print(time, DEC);
    Serial.print("    device:");
    Serial.print(device, DEC);
    Serial.print("    command:");
    Serial.println(command, DEC);
    #endif
    
    pAction new_action = (pAction)malloc(sizeof(tAction));
    new_action->time = time;
    new_action->device = device;
    new_action->data[0] = command; // data[0] contains the new state
    schedules.addElement(new_action);
}

// Removes a specific action from the actions list
// It does NOT remove it from flash memory
void removeSchedule(byte i)
{
    schedules.removeElement(i);
}

// Loads the current action list from the Arduino flash memory
void loadSchedules()
{
    byte number_actions;
    EEPROM_readAnything(EEPROM_ACTION_ADDRESS, number_actions);
    for(byte i=1; i<=number_actions; i++)
    {
        pAction new_action = (pAction)malloc(sizeof(tAction));
        EEPROM_readAnything(EEPROM_ACTION_ADDRESS+1+(i-1)*sizeof(tAction), *new_action);
        schedules.addElement(new_action);
    }        
}

// Saves the current action list in the Arduino flash memory
void saveSchedules()
{
    #ifdef DEBUG_ACTIONS
    Serial.println("Saving actions in Flash");
    #endif
    byte number_actions = schedules.count();

    EEPROM_writeAnything(EEPROM_ACTION_ADDRESS, number_actions);
    for(byte i=1; i<=number_actions; i++)
    {
        EEPROM_writeAnything(EEPROM_ACTION_ADDRESS+1+(i-1)*sizeof(tAction), *schedules.getPosition(i));
    }        
}

// Checks if it's time to process any new actions
void processSchedules(long previous_time, long time)
{
    // If the previous clock measurement was after 11:59:50 and the current one is before
    // 0:00:10, it means the day has changed, so we need to reset the scheduler
    if (previous_time > (CLOCK_MAX_DAY_SECONDS-10) && time < 10)
    {
        nextAction = 1;
    }    
    while (schedules.count()>0 && nextAction<=schedules.count() && time>=getActionTime(nextAction))
    {
        #ifdef DEBUG_ACTIONS
        PRINT_DEC(nextAction);
        Serial.print("time:");
        long t = getActionTime(nextAction);
        Serial.print(t);
        Serial.print("   dev:");
        Serial.print(getScheduleDevice(nextAction), DEC);
        Serial.print("   val:");
        Serial.println(schedules.getPosition(nextAction)->data[0], DEC);
        #endif
        setDeviceStatus(getScheduleDevice(nextAction), schedules.getPosition(nextAction)->data[0]);
        nextAction++;
    }
}

#endif




